import requests
import struct
import asyncio
import sys
import itertools
import re


DetectionMapping = {
    "clean": 0,
    "unknown": 1,
    "lib": 2,
    "malicious": 100,
}



def lookup_hashes(endpoint, service, hashes, license, ssl_verify=True):
    to_lookup = []
    for h in hashes:
        if type(h) == int:
            to_lookup.append(struct.pack("<Q", h).hex())
        elif type(h) == bytes:
            to_lookup.append(h.hex())
        else:
            raise ValueError(f"Invalid hash format: {h}")

    session = requests.Session()
    session.headers.update({'accept': 'application/json'})
    session.headers.update({'ApiKey': license})
    session.headers.update({'user-agent': "malcat"})
    r = session.post(f"{endpoint}/{service}", data={"hashes": ",".join(to_lookup), "retry": 1, "timeout": 30}, verify=ssl_verify)
    r.raise_for_status()
    data = r.json()
    status = data.get("status", "missing")
    if status != "ok":
        raise ValueError(f"Invalid job status {status}: {data.get('error', '')}")
    res = []
    for d in data.get("result", []):
        h =  bytes.fromhex(d["hash"])
        if len(h) == 8:
            h, = struct.unpack("<Q", h)
        lvl = DetectionMapping.get(d["level"], 1)
        res.append((h, lvl, d["score"], d["name"], d.get("paths", [])))
    return res



def get_stats(endpoint, license, ssl_verify=True):
    session = requests.Session()
    session.headers.update({'accept': 'application/json'})
    session.headers.update({'ApiKey': license})
    session.headers.update({'user-agent': "malcat"})
    r = session.get(f"{endpoint}/status", verify=ssl_verify)
    r.raise_for_status()
    data = r.json()
    stats = {}
    for worker in data.get("workers", []):
        v = worker.get("version", {})
        if not "Version" in v:
            continue
        if "Version" not in stats or stats["Version"] < v["Version"]:
            stats = v
    return list(stats.items())


def get_quota(endpoint, license, ssl_verify=True):
    session = requests.Session()
    session.headers.update({'accept': 'application/json'})
    session.headers.update({'ApiKey': license})
    session.headers.update({'user-agent': "malcat"})
    r = session.get(f"{endpoint}/quota", verify=ssl_verify)
    r.raise_for_status()
    data = r.json()
    return data.get("monthly_left", 0), data.get("monthly_total", 0)


def get_intel(endpoint, families, prefix, license, ssl_verify=True):
    session = requests.Session()
    session.headers.update({'accept': 'application/json'})
    session.headers.update({'ApiKey': license})
    session.headers.update({'user-agent': "malcat"})
    r = session.post(f"{endpoint}/families", data={"families": ",".join(families), "force_prefix": prefix.strip()}, verify=ssl_verify)
    r.raise_for_status()
    return r.json() or {}



def multi_lookup(endpoint, filetype, pic_hashes, string_hashes, constant_hash, license, ssl_verify=True):
    if sys.platform == "win32" and sys.version_info >= (3, 8, 0):
        asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())
    if endpoint.endswith("/"):
        endpoint = endpoint[:-1]
    # kesakode lookup
    async def multi_poll(pichashes, stringhashes, constant_hash, license, ssl_verify=True):
        loop = asyncio.get_event_loop()
        tasks = [
            loop.run_in_executor(None, lookup_hashes, endpoint, "pic", pichashes, license, ssl_verify),
            loop.run_in_executor(None, lookup_hashes, endpoint, "str", stringhashes, license, ssl_verify),
            loop.run_in_executor(None, lookup_hashes, endpoint, "cst", [constant_hash], license, ssl_verify),
            loop.run_in_executor(None, get_stats, endpoint, license, ssl_verify),
            loop.run_in_executor(None, get_quota, endpoint, license, ssl_verify),
        ]
        all_returns = await asyncio.gather(*tasks, return_exceptions=True)
        for r in all_returns:
            if isinstance(r, Exception):
                raise r
        return all_returns

    pic, str, cst, stats, quota = asyncio.run(multi_poll(pic_hashes, string_hashes, constant_hash, license, ssl_verify=ssl_verify))
    
    # gather intel
    if filetype == "PE":
        prefix = "win."
    elif filetype == "ELF":
        prefix = "elf."
    else:
        prefix = ""
    all_families = { x[3] for x in itertools.chain(pic, str, cst) if x[1] >= 100 }
    if all_families:
        intel = get_intel(endpoint, all_families, prefix, license, ssl_verify=ssl_verify)
    else:
        intel = {}
    return pic, str, cst, stats, intel, quota[0], quota[1]
    

def submit_sample(endpoint, file, license, is_malware=True, family="", comment=""):
    session = requests.Session()
    session.headers.update({'accept': 'application/json'})
    session.headers.update({'ApiKey': license})
    session.headers.update({'user-agent': "malcat"})
    r = session.post(f"{endpoint}/submit_sample", files={"file": file[:]}, data={"is_malware":is_malware, "malware_family": family, "comment": comment}, verify=False)
    r.raise_for_status()

